/*
 * Overclocking driver for Qualcomm MSM8960 (Krait) devices, such as the
 * HTC One XL (AT&T/Rogers/Asia) and HTC One S. This method is recommended for
 * rooted devices without kernel source or a locked bootloader.
 *
 * Successfully compiled against Code Aurora's ics_chocolate branch MSM kernel,
 * but requires a little tweaking to Module.symvers before it will successfully
 * load.
 *
 * Please encourage HTC and AT&T to allow the AT&T HTC One X's bootloader to be
 * unlocked! Even if it doesn't affect you, and even if it gets S-OFF
 * eventually, this sets a bad precedent for carriers to get in the way of
 * manufacturers' intentions of device freedom.
 * 
 * Copyright (c) 2012 Michael Huang
 * Author: Michael Huang <mike@setcpu.com>
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 */

#include <linux/cpufreq.h>
#include <linux/kallsyms.h>

#define DRIVER_AUTHOR "Michael Huang <mike@setcpu.com>"
#define DRIVER_DESCRIPTION "MSM 8960 Overclock Driver"
#define DRIVER_VERSION "1.0"

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESCRIPTION);
MODULE_VERSION(DRIVER_VERSION);
MODULE_LICENSE("GPL");

/* Speed of the HFPLL in KHz */
#define HFPLL_FREQ_KHZ 27000

/* Name of the acpu_freq_tbl symbol */
#define ACPU_FREQ_TBL_SYMBOL_NAME "acpu_freq_tbl_8960_kraitv2_nom"

/* Module parameters */
/* Controls overclocked frequency. 27 MHz * pll_l_val = X MHz */
static uint pll_l_val = 0x43;

/* Max currently supported is 1.3 V */
static uint vdd_uv = 1300000;

/* Symbol addresses. Unique per kernel version.
 *
 * Find the tables using:
 * "adb shell cat /proc/kallsyms | grep acpu_freq_tbl_8960_kraitv2"
 * This module defaults to using acpu_freq_tbl_8960_kraitv2_nom automatically.
 *
 * Look near the beginning the dmesg output to see which table your device
 * uses: nominal, slow, or fast.
 */
static ulong acpu_freq_tbl_addr = 0;

module_param(pll_l_val, uint, 0444);
module_param(vdd_uv, uint, 0444);
module_param(acpu_freq_tbl_addr, ulong, 0444);

static struct cpufreq_frequency_table freq_table[] = {
	{ 0, 384000 },
	{ 1, 486000 },
	{ 2, 594000 },
	{ 3, 702000 },
	{ 4, 810000 },
	{ 5, 918000 },
	{ 6, 1026000 },
	{ 7, 1134000 },
	{ 8, 1242000 },
	{ 9, 1350000 },
	{ 10, 1458000 },
	{ 11, 1512000 },
	{ 12, 1809000 }, /* We replace this with our desired speed later */
	{ 13, CPUFREQ_TABLE_END },
};

struct core_speed {
	unsigned int		khz;
	int			src;
	unsigned int		pri_src_sel;
	unsigned int		sec_src_sel;
	unsigned int		pll_l_val;
};

struct acpu_level {
	unsigned int		use_for_scaling;
	struct core_speed	speed;
	struct l2_level		*l2_level;
	unsigned int		vdd_core;
};

struct cpufreq_frequency_table *orig_table;

/* Use a function pointer for cpufreq_cpu_get because the symbol version
 * differs in the HTC kernel and the Code Aurora kernel, so the kernel won't
 * let us call it normally.
 *
 * Call the function normally when the kernel source is released.
 */
typedef struct cpufreq_policy *(*cpufreq_cpu_get_type)(int); 
struct cpufreq_policy *(*cpufreq_cpu_get_new)(int);

static int __init overclock_init(void)
{
	struct cpufreq_policy *policy;
	struct acpu_level *acpu_freq_tbl;
	uint cpu;

	printk(KERN_INFO "krait_oc: %s version %s\n", DRIVER_DESCRIPTION,
		DRIVER_VERSION);
	printk(KERN_INFO "krait_oc: by %s\n", DRIVER_AUTHOR);
	printk(KERN_INFO "krait_oc: overclocking to %u at %u uV\n",
		pll_l_val*HFPLL_FREQ_KHZ, vdd_uv);

	cpufreq_cpu_get_new =
		(cpufreq_cpu_get_type) kallsyms_lookup_name("cpufreq_cpu_get");

	printk(KERN_INFO "krait_oc: updating cpufreq policy\n");
	policy = cpufreq_cpu_get_new(0);
	policy->cpuinfo.max_freq = pll_l_val*HFPLL_FREQ_KHZ;

	printk(KERN_INFO "krait_oc: updating cpufreq frequency tables\n");
	freq_table[12].frequency = pll_l_val*HFPLL_FREQ_KHZ;

	/* Save a pointer to the freq original table to restore if unloaded */
	orig_table = cpufreq_frequency_get_table(0);

	for_each_possible_cpu(cpu) {
		cpufreq_frequency_table_put_attr(cpu);
		cpufreq_frequency_table_get_attr(freq_table, cpu);
	}

	printk(KERN_INFO "krait_oc: updating acpu_freq_tbl\n");
	if(acpu_freq_tbl_addr == 0)
		acpu_freq_tbl_addr =
		kallsyms_lookup_name(ACPU_FREQ_TBL_SYMBOL_NAME);

	acpu_freq_tbl = (struct acpu_level*) acpu_freq_tbl_addr;

	/* Index 20 is not used for scaling in the acpu_freq_tbl, so fill it
         * with our new freq */
	acpu_freq_tbl[20].speed.khz = pll_l_val*HFPLL_FREQ_KHZ;
	acpu_freq_tbl[20].speed.pll_l_val = pll_l_val;
	acpu_freq_tbl[20].vdd_core = vdd_uv;

	return 0;
}

static void __exit overclock_exit(void)
{
	struct cpufreq_policy *policy;
	uint cpu;

	printk(KERN_INFO "krait_oc: reverting cpufreq policy\n");
	policy = cpufreq_cpu_get_new(0);
	policy->cpuinfo.max_freq = 1512000;

	printk(KERN_INFO "krait_oc: reverting cpufreq frequency tables\n");
	for_each_possible_cpu(cpu) {
		cpufreq_frequency_table_put_attr(cpu);
		cpufreq_frequency_table_get_attr(orig_table, cpu);
	}

	printk(KERN_INFO "krait_oc: unloaded\n");
}

module_init(overclock_init);
module_exit(overclock_exit);
